from typing import Dict, Any


class ConfigScope:
    pass


def load_config_from_file(path: str) -> Dict[str, Any]:
    import json
    import yaml
    with open(path) as fin:
        if path.endswith('.yml'):
            return yaml.load(fin)
        elif path.endswith('.json'):
            return json.load(fin)


def load_config_in_workdirectory(filename):
    pass


def load_config_in_user_config(filename, default_root=None):
    pass


"""
c = cfg.get_module_config(__name__)
with cfg.ConfigScope():

"""



class QueryKey:
    """
    <Immutable>
    """

    def __init__(self, key=None):
        if isinstance(key, QueryKey):
            self.data = key.data
        elif isinstance(key, str):
            self.data = self.parse_str(key)
        elif key is None:
            self.data = ()
        elif isinstance(key, (list, tuple)):
            self.data = tuple(key)
        else:
            raise TypeError(
                "Key {} is not convertable to QueryKey.".format(key))

    def head(self):
        if len(self.data) == 0:
            return None
        return self.data[0]

    def tail(self):
        return QueryKey(self.data[1:])

    def last(self):
        if len(self.data) == 0:
            return None
        return self.data[-1]

    def __len__(self):
        return len(self.data)

    def __eq__(self, q):
        return self.data == QueryKey(q).data

    def get_query(self, dct):
        for k in self.data:
            dct = dct[k]
        return dct

    def set_query(self, dct, value):
        if len(self.data) < 1:
            raise ValueError("Set query requires len(key) >= 1.")
        for k in self.data[:-1]:
            v = dct.get(k)
            if v is None:
                dct[k] = {}
            dct = v
        dct[self.data[-1]] = value

    @classmethod
    def parse_str(self, key):
        return key.split('/')


import abc
import threading
from ..config import cnode, view


class Sourceable(object):

	__metaclass__ = abc.ABCMeta

	@abc.abstractmethod
	def read(self, keys):
		"""read Cnode/Cview from a configuration tree ."""
		return

	@abc.abstractmethod
	def write(self, keys, config_obj, is_overwrite=False):
		"""add new configuration to a singleton config tree"""
		return


class AbstractProxy(object):

	__metaclass__ = abc.ABCMeta

	@abc.abstractmethod
	def create_proxy(self):
		""" factory create method."""
		return


class Configuration(Sourceable):

	__root = None

	def __init__(self, root_node):
		if not isinstance(root_node, cnode.CNode):
			raise ValueError("root must be Cnode object.")
		self.set_root(root_node)

	def set_root(self, root):
		self.__root = root

	def get_root(self):
		return self.__root

	def read(self, keys)-> view.CView:
		if self.__root is not None:
			cv = view.CView(self.__root)
			return cv.get(keys)

	def __getitem__(self, keys):
		return self.read(keys)

	def write(self, keys, config_obj: [str, dict, cnode.CNode], is_overwrite=False):
		self.__root.update(keys, config_obj, overwrite_node=is_overwrite)


class ConfigProxy(AbstractProxy):

	"""a proxy for configuration."""
	_instance_lock = threading.Lock()
	__configs = {}
	_instance = None

	def __init__(self, root=None, root_id=None):
		self.init_params = [root, root_id]
		if self.init_params[0] is not None and self.init_params[1] is not None:
			self.create_proxy()

	def __new__(cls, *args, **kwargs):
		if ConfigProxy._instance is None:
			with ConfigProxy._instance_lock:
				if ConfigProxy._instance is None:
					ConfigProxy._instance = object.__new__(cls)
		return ConfigProxy._instance

	@classmethod
	def reset(cls):
		if ConfigProxy._instance is not None:
			ConfigProxy._instance = None
			ConfigProxy.__configs = {}

	def create_proxy(self):
		return self.add_proxy(self.init_params[0], self.init_params[1])

	def add_proxy(self, config_obj: Configuration, config_id: str):
		"""assign a name(ID) for every config"""
		if not isinstance(config_obj, Configuration):
			raise ValueError('config_obj must be a Configuration object.')
		if config_id in self.__configs.keys():
			raise ValueError('config_id already exists.')
		self.__configs[config_id] = config_obj

	def write(self, config_id, keys, config_obj, is_overwrite=False):
		return self.__configs[config_id].write(keys, config_obj, is_overwrite=is_overwrite)

	def read(self, config_id, config_key):
		return self.__configs[config_id].read(config_key)

	def __getitem__(self, item):
		return self.__configs[item]

	def __setitem__(self, config_id, root):
		if config_id in self.__configs.keys():
			if isinstance(root, Configuration):
				self.__configs[config_id] = root
				return
			else:
				raise ValueError('item must be a configuration object.')
		self.add_proxy(root, config_id)
		return

	def get_root_node(self, config_id):
		"""get configuration's root Cnode corresponding to config_id"""
		if config_id not in self.__configs.keys():
			return None
		return self.__configs[config_id].get_root()

	def set_root_node(self, config_id, root):
		"""reset configuration's root Cnode corresponding to config_id"""
		if config_id not in self.__configs.keys():
			raise ValueError('configuration not found .')
		return self.__configs[config_id].set_root(root)





"""
Useage:

1. load config to a dict object
2. following these codes:

```
dct = {
    'aaa':{
        'bbb':{
            'x': 1,
            'y': 2,
        }
        'z' : 3,
    }
}
c = create_config_tree()
v = create_view(c, 'aaa/bbb')
v['x']
>>> 1
v['y']
>>> 2
v['z']
>>> 3
```

"""
"""
Config module make it easy for config intens projects.

Config module consists of following classes:
ConfigsViewer:
    Configs viewer class.

ConfigsMaker:
    

"""
"""
New design:
    This package should consisted of the following classes:

    1.  Viewer (R)
        A Viewer is a object with 
        Provide facility for querying config via:
            normal key/name: `'key'`
            path-like: `'aaa/bbb/ccc'` and `'aaa.bbb.ccc'` (configurable)
    
    2.  Loader (C)
        Construct config tree from files/scripts
    
    3.  Updator (U)
        Update config tree
    
    4.  CNode / CTree (Model)
        Data structure of storing configs in memory
        Support serilization to .json/.yaml

        To deal with multiple config tree combination (multiple default):
            CNodeAnonymous
    
    Also provide the following helper functions:
    default() -> default config tree

Design V3:
    1. Everything is saved in a real dict (no any special objects).
    2. All operations were done via CNode.
    3. Basic views were done via CView.
    4. Advanced view were done by concrete class of CSpec.

    class CNode:
        def __init__(self, data:dict):
            pass
        
        def __getitem__(self, key):
            pass

"""

# from .base import Configs
# from .module_config import ModuleConfigs
# from ._viewer import ConfigsView
# from ._configurable import configurable

from .cnode import from_dict as create_config_tree
create_node = create_config_tree
from .cnode import CNode
from .view import CView
from .view import create_view


from typing import Dict
from enum import Enum


from .query_key import QueryKey


class Keywords:
    EXPAND = '__expand__'


def need_expand(v):
    if not isinstance(v, dict):
        return False
    if Keywords.EXPAND in v:
        return v[Keywords.EXPAND]
    return True


from collections import UserDict


class CNode(UserDict):
    class KEYS:
        NO_EXPAND = '_no_expand_'

    def __init__(self, data=None, father=None):
        """
        Args:
            `data`: dict or CNode.
        """
        self.data = {}
        if data is None:
            data = {}
        if isinstance(data, dict):
            for k, v in data.items():
                self._update_kernel(k, v)
        elif isinstance(data, CNode):
            self.data = data.data
        else:
            raise TypeError("data {} is not convertable")
        self.father = father

    def _update_kernel(self, key, value):
        if not isinstance(value, dict) or value.get(self.KEYS.NO_EXPAND):
            self.data[key] = value
        else:
            self.data[key] = CNode(value, self)

    def __setitem__(self, key, value):
        self._update_kernel(key, value)

    def read(self, key: QueryKey, default_value=None):
        """
        Find value curresponding to key.
        if len(key) == 1, then find it in self._values,
        else find it in self._children recuresivly.
        """
        key = QueryKey(key)
        if len(key) == 0:
            return self
        if len(key) == 1:
            return self.get(key.head())
        if not key.head() in self.data or not isinstance(
                self.data[key.head()], CNode):
            return default_value
        return self.data.get(key.head()).read(key.tail())

    @property
    def children(self):
        return {k: v for k, v in self.data.items() if isinstance(v, CNode)}

    def create(self, key: QueryKey, node_or_value):
        """
        Create a new child node or value.
        """
        return self.update(key, node_or_value, allow_exist=False)

    def is_ancestor_of(self, n):
        for k, v in self.data.items():
            if not isinstance(v, CNode):
                continue
            if v is n or v.is_ancestor_of(n):
                return True
        return False

    def update(self,
               key: QueryKey,
               node_or_value,
               *,
               allow_exist=True,
               overwrite_node=False):
        """
        Updating config.
        If node_or_value is a value, update directly.
        If node_or_value is a CNode, check if this node exists.
            If not exists: direct assign.
            If exists: update each item of that node. 
        Note: the node_or_value argument is not assigned to the node tree.
        """
        key = QueryKey(key)
        if len(key) == 0:
            if not isinstance(node_or_value, (dict, CNode)):
                raise ValueError(
                    "Can not update node with none CNode object or dict, overrite_node might need to be True."
                )
            for k, v in node_or_value.items():
                if isinstance(self.data.get(k), CNode):
                    self.data[k].update(v)
                else:
                    self.data[k] = v
            return self
        if key.head() in self.data and not allow_exist:
            raise ValueError("Key {} alread existed.".format(key.head()))
        if len(key) > 1:
            if not isinstance(self.data.get(key.head()),
                              CNode) or overwrite_node:
                self.data[key.head()] = CNode().update(key.tail(),
                                                       node_or_value)
            else:
                self.data[key.head()].update(key.tail(), node_or_value)
            return self
        if len(key) == 1:
            if not isinstance(self.data.get(key.head()),
                              CNode) or overwrite_node:
                if need_expand(node_or_value):
                    self.data[key.head()] = from_dict(node_or_value)
                else:
                    self.data[key.head()] = node_or_value
            else:
                self.data[key.head()].update([], node_or_value)
            return self


def from_dict(config_dict):
    return CNode(config_dict)


class DefaultConfig:
    _current = None

    def __init__(self, cnode):
        pass

    @property
    def node(self):
        return self._current

    def __enter__(self):
        pass


class CNodeRoot(CNode):
    pass



import logging


class SimpleInMemoryBackend:
    """
    Simple in memory backend, designed to simplify tests.
    """

    def __init__(self):
        self.messages = []

    def format(self, level, message):
        def level_name(level):
            if level == logging.INFO:
                return 'INFO'
            raise ValueError("Unknown level.")

        return '[{}] {}'.format(level_name(level), message)

    def info(self, message):
        self.messages.append(self.format(logging.INFO, message))

    def content(self):
        return '\n'.join(self.messages)

    def clear(self):
        self.messages = []

    def __len__(self):
        """
        Returns number of messages added.
        """
        return len(self.messages)


# Alias of logging
StandardBackend = logging


from .view import CView


class Specs(CView):
    FIELDS = tuple()

    # def __init__(self, root, path):
    #     self.data = {k: v for k, v in config.items()
    #                  if k in self.FIELDS}

    def __getattr__(self, key):
        if key in self.FIELDS:
            return self.data[key]
        raise KeyError("Key {} not found.".format(key))



"""
dxl.core.logger utilities is providing a AOP 范式 logger, 
based on standard logging system.

Objective:
- clear code base with AOP
- advantages from standard logging system
"""
from ._logger import Logger
from .backend import SimpleInMemoryBackend



import json
import yaml
from typing import Dict, Any

def load_config(path: str) -> Dict[str, Any]:
    with open(path) as fin:
        if path.endswith('.yml') or path.endswith('.yaml'):
            return yaml.load(fin)
        if path.endswith('.json'):
            return json.load(fin)



import time
import datetime

PROGRESS_FORMATS = {
    'std': "[{current:>8}/{total:>8}({complete_percentage:5.2f}%, {iteration_speed}) | {time_elapsed} < {time_estimate}]"
}

MIN_ELAPSED_TIME = 1e-5


class ProgressReporter:
    def __init__(self, total=100, message=None, min_elapsed_time=1.0, print_func=None):
        self.total = total
        self.min_elapsed_time = min_elapsed_time
        self.message = message or PROGRESS_FORMATS['std']
        self.step = 0
        self.time_start = None
        self.time_previous = None
        self.print_func = print_func if print_func is not None else print
        self.reset()

    def reset(self):
        self.time_start = time.time()
        self.time_next_report = self.time_start

    def set_time_next_report(self, previous):
        self.time_next_report = previous + self.min_elapsed_time

    def complete_percentage(self):
        return float(self.step) / float(self.total) * 100

    def time_estimate(self):
        if self.complete_percentage() <= 0.0:
            return None
        return (1 / self.complete_percentage() * 100 - 1) * self.time_elapsed()

    def time_elapsed(self):
        return time.time() - self.time_start

    def iterations_per_second(self):
        return self.step / max(self.time_elapsed(), MIN_ELAPSED_TIME)

    def iteration_speed(self) -> str:
        if self.iterations_per_second() > 1.0 or self.iterations_per_second() < MIN_ELAPSED_TIME:
            return "{:4.2f}it/s".format(self.iterations_per_second())
        else:
            return "{:4.2f}s/it".format(1.0 / self.iterations_per_second())

    def passed_enough_time(self):
        return time.time() > self.time_next_report

    def update(self, step=None, *args, **kwargs):
        if step is None:
            step = self.step + 1
        self.step = step

        if not self.passed_enough_time():
            return

        self.print_func(self.message.format(*args,
                                            current=self.step,
                                            total=self.total,
                                            complete_percentage=self.complete_percentage(),
                                            iteration_speed=self.iteration_speed(),
                                            time_elapsed=str(datetime.timedelta(
                                                seconds=self.time_elapsed())),
                                            time_estimate=str(datetime.timedelta(
                                                seconds=self.time_estimate())),
                                            **kwargs))
        self.set_time_next_report(time.time())



import cProfile
from io import StringIO
import pstats
import contextlib


@contextlib.contextmanager
def profiled():
    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    s = StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
    ps.print_stats()
    # uncomment this to see who's calling what
    # ps.print_callers()
    print(s.getvalue())



from functools import wraps
class LoggerAspect:
    def __init__(self, backend):
        self.backend = backend

    def make_message(self, message, *args, **kwargs):
        return str(message).format(*args, **kwargs)

    def make_log(self, log_maker, message):
        raise NotImplementedError

    def info(self, message):
        return self.make_log(self.backend.info, message)

    def debug(self, message):
        return self.make_log(self.backend.debug, message)

    def error(self, message):
        return self.make_log(self.backend.error, message)


class LoggerBefore(LoggerAspect):
    def make_log(self, log_maker, message):
        def advice(func):
            @wraps(func)
            def call(*args, **kwargs):
                log_maker(self.make_message(message, *args, **kwargs))
                return func(*args, **kwargs)

            return call

        return advice


class LoggerAfter(LoggerAspect):
    def make_log(self, log_maker, message):
        def advice(func):
            @wraps(func)
            def call(*args, **kwargs):
                result = func(*args, **kwargs)
                if isinstance(result, dict):
                    log_maker(self.make_message(message, **result))
                elif isinstance(result, (list, tuple)):
                    log_maker(self.make_message(message, *result))
                else:
                    log_maker(self.make_message(message, result))
                return result

            return call

        return advice


from pathlib import Path

DEFAULT_PATH = '/gluster/share/develop/fixtures'


class PersistentFixturesManager:
    """
    Singleton. 
    """
    _path = None

    @classmethod
    def path(cls):
        """
        Root path of all fixture data.
        """
        return Path(cls._path)

    @classmethod
    def default_path(cls):
        return Path(DEFAULT_PATH)

    @classmethod
    def reset(cls):
        """
        Reset configs to its default value.
        """
        cls._path = cls.default_path()



from .aspect import LoggerBefore, LoggerAfter
import logging


class Logger:
    def __init__(self, name_or_backend):
        """
        If name_or_backend is str, 
        use default backend logging.getLogger(name_or_backend).

        Backend should have methods:
        `b.info`, `b.debug`, `b.error`, `b.warning`, etc.
        """
        if isinstance(name_or_backend, str):
            self.backend = logging.getLogger(name_or_backend)
        else:
            self.backend = name_or_backend
        self.before_aspect = LoggerBefore(self.backend)
        self.after_aspect = LoggerAfter(self.backend)

    @property
    def before(self):
        return self.before_aspect

    @property
    def after(self):
        return self.after_aspect

    def info(self, message):
        self.backend.info(message)



import sys


class ExceptionHook:
  instance = None

  def __call__(self, *args, **kwargs):
    if self.instance is None:
      from IPython.core import ultratb
      self.instance = ultratb.FormattedTB(
          mode='Plain', color_scheme='Linux', call_pdb=1)
    return self.instance(*args, **kwargs)


def enter_debug():
  sys.excepthook = ExceptionHook()



from setuptools import setup, find_packages
setup(
    name='dxl-core',
    version='0.1.4',
    description='Core utility library.',
    url='https://github.com/Hong-Xiang/dxcore',
    author='Hong Xiang',
    author_email='hx.hongxiang@gmail.com',
    license='MIT',
    namespace_packages=['dxl'],
    packages=find_packages('src/python'),
    package_dir={'': 'src/python'},
    install_requires=[],
    zip_safe=False)



import logging


def set_logging_info_level_with_default_format():
    logging.basicConfig(
        level=logging.INFO,
        format=
        # '%(asctime)s %(filename)s [line:%(lineno)d] %(levelname)s %(message)s',
        '%(asctime)s [%(levelname)s] %(message)s',
        # datefmt='%a, %d %b %Y %H:%M:%S')
        datefmt='%d %b %Y %H:%M:%S')



import logging

logging.basicConfig(
    format='[%(levelname)s] %(asctime)s [%(filename)s:%(lineno)d] %(message)s',
    datefmt='%a, %d %b %Y %H:%M:%S',
)
logger = logging.getLogger('dxl')


def set_logging_level(level=logging.INFO):
    logger.setLevel(level)


__all__ = ['logger', 'set_logging_level']



#from dxl.fs import Path

from pathlib import Path
from .cnode import QueryKey, CNode


def _parse_query_key(key: str, parse_dot=True):
    if parse_dot:
        for i, v in enumerate(key):
            if i == 0:
                continue
            if i == len(key) - 1:
                continue
            if v != '.':
                continue
            if key[i - 1] == '/' or key[i + 1] == '/':
                continue
            key[i] = '/'
    kp = Path(key)
    if str(kp) == '/':
        return QueryKey([])
    return QueryKey(kp.parts)


def _find_ancestors(root, cnode):
    result = []
    current = root
    found = True
    while current.is_ancestor_of(cnode):
        result.append(current)
        found = False
        for _, v in current.children.items():
            if v.is_ancestor_of(cnode):
                current = v
                found = True
                break
        if not found:
            break
    return tuple(result)


def base_node(root, key_path):
    result = root.read(_parse_query_key(key_path))
    if result is None:
        raise ValueError("No node of {} is found.".format(key_path))
        # root.update(_parse_query_key(key_path), CNode())
        # result = root.read(_parse_query_key(key_path))
    return result


class CView:
    """
    Viewer of CNode.
    Provide simplified hierarchy access to values.
    """

    def __init__(self, root, base=None):
        if isinstance(root, dict):
            root = CNode(root)
            base = root
        elif isinstance(root, CView):
            root = root.root
            if base is None:
                base = root.base

        self._root = root
        if base is None:
            self._base = root
        else:
            self._base = base

    @property
    def root(self):
        return self._root

    @property
    def base(self):
        return self._base

    def get_kernel(self, key, value=None):
        v = self._base.read(_parse_query_key(key))
        if isinstance(v, CNode):
            return CView(self._root, v)
        if v is None:
            return value
        return v

    def search(self, key):
        v = self.get_kernel(key)
        if v is not None:
            return v
        acs = _find_ancestors(self._root, self._base)
        for n in reversed(acs):
            v = CView(self._root, n).search(key)
            if v is not None:
                return v
        return v

    def __getitem__(self, key):
        return self.search(key)

    def get(self, key, value=None):
        v = self.search(key)
        if v is None:
            return value
        return v

    def items(self):
        return self._base.items()


def create_view(root, key_path):
    """
    Get config view by path
    """
    return CView(root, base_node(root, key_path))


# class ConfigViewer:
#     def __unified_keys(self, path_or_keys):
#         if isinstance(path_or_keys, (list, tuple)):
#             return list(self.base.parts()), list(path_or_keys)
#         else:
#             return list(self.base.parts()), list(Path(path_or_keys).parts())

#     def _get_value_raw(self, keys):
#         result = self.data
#         for k in keys:
#             if not isinstance(result, (dict, ConfigsView)):
#                 return None
#             result = result.get(k)
#         return result

#     def _form_path(self, keys):
#         return '/'.join(keys)

#     def _query(self, base_keys, local_keys):
#         # if len(local_keys) <= 1:
#         result = self._get_value_raw(base_keys + local_keys)
#         # else:
#         # path = self.base / local_keys[0]
#         # key_path = self._form_path(local_keys[1:])
#         # result = ConfigsView(self.data, path)._query(key_path)
#         return result

#     def _search(self, key):
#         base_keys, local_keys = self.__unified_keys(key)
#         result = self._query(base_keys, local_keys)
#         path = self._form_path(base_keys + local_keys)
#         if result is None and len(base_keys) > 0:
#             new_path = self._form_path((base_keys + local_keys)[:-2])
#             result, _ = ConfigsView(self.data, new_path)._search(
#                 local_keys[-1])
#         return result, path

#     def _post_processing(self, result, path, default, restrict):
#         if isinstance(result, dict) or (result is None and default is None
#                                         and not restrict):
#             result = ConfigsView(self.data, path)
#         elif result is None:
#             result = default
#         return result

#     def to_dict(self):
#         result, path = self._search(str(self.base))
#         if isinstance(result, dict):
#             return result
#         else:
#             raise ValueError("Configs view on base path is not dict.")

#     def get(self, key, default=None, *, restrict=False):
#         result, path = self._search(key)
#         return self._post_processing(result, path, default, restrict)

#     def __getitem__(self, key):
#         return self.get(key, restrict=True)

#     def __iter__(self):
#         dct = self._get_value_raw(self.base.parts())
#         if dct is None:
#             return list().__iter__()
#         else:
#             return dct.__iter__()

# def view(path_like_key, config=None):
#     pass

# def child_view(configs_view, extend_path):
#     return ConfigsView(configs_view.base_path / str(extend_path),
#                        configs_view.data)
